# @file dec_parser.py
# Code to help parse DEC file
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser
import os


class DecParser(HashFileParser):
    def __init__(self):
        HashFileParser.__init__(self, 'DecParser')
        self.Lines = []
        self.Parsed = False
        self.Dict = {}
        self.LibrariesUsed = []
        self.PPIsUsed = []
        self.ProtocolsUsed = []
        self.GuidsUsed = []
        self.PcdsUsed = []
        self.IncludesUsed = []
        self.Path = ""

    def ParseFile(self, filepath):
        self.Logger.debug("Parsing file: %s" % filepath)
        if(not os.path.isabs(filepath)):
            fp = self.FindPath(filepath)
        else:
            fp = filepath
        self.Path = fp

        f = open(fp, "r")
        self.Lines = f.readlines()
        f.close()
        InDefinesSection = False
        InLibraryClassSection = False
        InProtocolsSection = False
        InGuidsSection = False
        InPPISection = False
        InPcdSection = False
        InIncludesSection = False

        for line in self.Lines:
            sline = self.StripComment(line)

            if(sline is None or len(sline) < 1):
                continue

            if InDefinesSection:
                if sline.strip()[0] == '[':
                    InDefinesSection = False
                else:
                    if sline.count("=") == 1:
                        tokens = sline.split('=', 1)
                        self.Dict[tokens[0].strip()] = tokens[1].strip()
                        continue

            elif InLibraryClassSection:
                if sline.strip()[0] == '[':
                    InLibraryClassSection = False
                else:
                    t = sline.partition("|")
                    self.LibrariesUsed.append(t[0].strip())
                    continue

            elif InProtocolsSection:
                if sline.strip()[0] == '[':
                    InProtocolsSection = False
                else:
                    t = sline.partition("=")
                    self.ProtocolsUsed.append(t[0].strip())
                    continue

            elif InGuidsSection:
                if sline.strip()[0] == '[':
                    InGuidsSection = False
                else:
                    t = sline.partition("=")
                    self.GuidsUsed.append(t[0].strip())
                    continue

            elif InPcdSection:
                if sline.strip()[0] == '[':
                    InPcdSection = False
                else:
                    t = sline.partition("|")
                    self.PcdsUsed.append(t[0].strip())
                    continue

            elif InIncludesSection:
                if sline.strip()[0] == '[':
                    InIncludesSection = False
                else:
                    self.IncludesUsed.append(sline.strip())
                    continue

            elif InPPISection:
                if (sline.strip()[0] == '['):
                    InPPISection = False
                else:
                    t = sline.partition("=")
                    self.PPIsUsed.append(t[0].strip())
                    continue

            # check for different sections
            if sline.strip().lower().startswith('[defines'):
                InDefinesSection = True

            elif sline.strip().lower().startswith('[libraryclasses'):
                InLibraryClassSection = True

            elif sline.strip().lower().startswith('[protocols'):
                InProtocolsSection = True

            elif sline.strip().lower().startswith('[guids'):
                InGuidsSection = True

            elif sline.strip().lower().startswith('[ppis'):
                InPPISection = True

            elif sline.strip().lower().startswith('[pcd'):
                InPcdSection = True

            elif sline.strip().lower().startswith('[includes'):
                InIncludesSection = True

        self.Parsed = True



##
# UEFI Authenticated Variable Structure Support Library
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import logging
import datetime
import struct
import hashlib
import uuid
from edk2toollib.uefi.wincert import WinCert, WinCertUefiGuid
from edk2toollib.utility_functions import PrintByteList


'''
Structures definition based on UEFI specification (UEFI 2.7)

Each object can be created and or populated from a file stream.
Each object can be written to a filesteam as binary and printed to the console in text.
'''

# UEFI global Variable Namespace
EfiGlobalVarNamespaceUuid = uuid.UUID('8BE4DF61-93CA-11d2-AA0D-00E098032B8C')
Sha256Oid = [0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01]

#
# EFI_SIGNATURE_DATA Structure for X509 Certs
#


class EfiSignatureDataEfiCertX509(object):
    STATIC_STRUCT_SIZE = 16

    #
    # decodefs is a filestream object of binary content that is the structure encoded
    # decodesize is number of bytes to decode as the EFI_SIGNATURE_DATA object (guid + x509 data)
    # createfs is a filestream object that is the DER encoded x509 cert
    # sigowner is the uuid object of the signature owner guid
    def __init__(self, decodefs=None, decodesize=0, createfs=None, sigowner=None):
        if(decodefs is not None):
            self.PopulateFromFileStream(decodefs, decodesize)
        elif(createfs is not None):
            # create a new one
            self.SignatureOwner = sigowner
            start = createfs.tell()  # should be 0 but maybe this filestream has other things at the head
            createfs.seek(0, 2)
            end = createfs.tell()
            createfs.seek(start)
            self.SignatureDataSize = end - start
            if(self.SignatureDataSize < 0):
                raise Exception("Create File Stream has invalid size")
            self.SignatureData = memoryview(createfs.read(self.SignatureDataSize))

        else:
            raise Exception("Invalid Parameters - Not Supported")

    def PopulateFromFileStream(self, fs, decodesize):
        if(fs is None):
            raise Exception("Invalid File Steam")

        if(decodesize == 0):
            raise Exception("Invalid Decode Size")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if((end - offset) < EfiSignatureDataEfiCertX509.STATIC_STRUCT_SIZE):  # size of the  guid
            raise Exception("Invalid file stream size")

        if((end - offset) < decodesize):  # size requested is too big
            raise Exception("Invalid file stream size vs decodesize")

        self.SignatureOwner = uuid.UUID(bytes_le=fs.read(16))

        # read remainling decode size for x509 data
        self.SignatureDataSize = decodesize - EfiSignatureDataEfiCertX509.STATIC_STRUCT_SIZE
        self.SignatureData = memoryview(fs.read(self.SignatureDataSize))

    def Print(self):
        print("EfiSignatureData - EfiSignatureDataEfiCertX509")
        print("  Signature Owner:      %s" % str(self.SignatureOwner))
        print("  Signature Data: ")
        if(self.SignatureData is None):
            print("    NONE")
        else:
            sdl = self.SignatureData.tolist()
            if(self.SignatureDataSize != len(sdl)):
                raise Exception("Invalid Signature Data Size vs Length of data")
            PrintByteList(sdl)

    def Write(self, fs):
        if(fs is None):
            raise Exception("Invalid File Output Stream")
        if(self.SignatureData is None):
            raise Exception("Invalid object")

        fs.write(self.SignatureOwner.bytes_le)
        fs.write(self.SignatureData)

    def GetTotalSize(self):
        return EfiSignatureDataEfiCertX509.STATIC_STRUCT_SIZE + self.SignatureDataSize


#
# EFI_SIGNATURE_DATA Structure for Sha256 hash
#
class EfiSignatureDataEfiCertSha256(object):
    STATIC_STRUCT_SIZE = 16 + hashlib.sha256().digest_size  # has guid and array

    #
    # decodefs is a filestream object of binary content that is the structure encoded
    # createfs is a filestream object of binary that is to be hashed to create the signature data
    # digest is a byte array that contains the hash value for new signature data
    # sigowner is the uuid object of the signature owner guid
    def __init__(self, decodefs=None, createfs=None, digest=None, sigowner=None):
        if(decodefs is not None):
            self.PopulateFromFileStream(decodefs)
        elif(createfs is not None):
            # create a new one
            self.SignatureOwner = sigowner
            self.SignatureData = memoryview(hashlib.sha256(createfs.read()).digest())
        elif(digest is not None):
            self.SignatureOwner = uuid.UUID(sigowner)
            self.SignatureData = memoryview(digest)
        else:
            raise Exception("Invalid Parameters - Not Supported")

    def PopulateFromFileStream(self, fs):
        if(fs is None):
            raise Exception("Invalid File Steam")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if((end - offset) < EfiSignatureDataEfiCertSha256.STATIC_STRUCT_SIZE):  # size of the  data
            raise Exception("Invalid file stream size")

        self.SignatureOwner = uuid.UUID(bytes_le=fs.read(16))

        self.SignatureData = memoryview(fs.read(hashlib.sha256().digest_size))

    def Print(self):
        print("EfiSignatureData - EfiSignatureDataEfiCertSha256")
        print("  Signature Owner:      %s" % str(self.SignatureOwner))
        print("  Signature Data: ", end="")
        if(self.SignatureData is None):
            print(" NONE")
        else:
            sdl = self.SignatureData.tolist()
            for index in range(len(sdl)):
                print("%02X" % sdl[index], end='')
            print("")

    def Write(self, fs):
        if(fs is None):
            raise Exception("Invalid File Output Stream")
        if(self.SignatureData is None):
            raise Exception("Invalid object")

        fs.write(self.SignatureOwner.bytes_le)
        fs.write(self.SignatureData)

    def GetTotalSize(self):
        return EfiSignatureDataEfiCertSha256.STATIC_STRUCT_SIZE


class EfiSignatureHeader(object):
    def __init__(self):
        raise Exception("Not Implemented")


class EfiSignatureDataFactory(object):
    EFI_CERT_SHA256_GUID = uuid.UUID('c1c41626-504c-4092-aca9-41f936934328')
    # EFI_CERT_RSA2048_GUID = uuid.UUID("0x3c5766e8, 0x269c, 0x4e34, 0xaa, 0x14, 0xed, 0x77, 0x6e, 0x85, 0xb3, 0xb6")
    # EFI_CERT_RSA2048_SHA256_GUID = uuid.UUID("0xe2b36190, 0x879b, 0x4a3d, 0xad, 0x8d, 0xf2, 0xe7, 0xbb, 0xa3, 0x27, 0x84")  # noqa: E501
    # EFI_CERT_SHA1_GUID = uuid.UUID("0x826ca512, 0xcf10, 0x4ac9, 0xb1, 0x87, 0xbe, 0x1, 0x49, 0x66, 0x31, 0xbd")
    # EFI_CERT_RSA2048_SHA1_GUID = uuid.UUID("0x67f8444f, 0x8743, 0x48f1, 0xa3, 0x28, 0x1e, 0xaa, 0xb8, 0x73, 0x60, 0x80")    # noqa: E501
    EFI_CERT_X509_GUID = uuid.UUID("a5c059a1-94e4-4aa7-87b5-ab155c2bf072")
    # EFI_CERT_SHA224_GUID = uuid.UUID("0xb6e5233, 0xa65c, 0x44c9, 0x94, 0x7, 0xd9, 0xab, 0x83, 0xbf, 0xc8, 0xbd")
    # EFI_CERT_SHA384_GUID = uuid.UUID("0xff3e5307, 0x9fd0, 0x48c9, 0x85, 0xf1, 0x8a, 0xd5, 0x6c, 0x70, 0x1e, 0x1")
    # EFI_CERT_SHA512_GUID = uuid.UUID("0x93e0fae, 0xa6c4, 0x4f50, 0x9f, 0x1b, 0xd4, 0x1e, 0x2b, 0x89, 0xc1, 0x9a")
    EFI_CERT_X509_SHA256_GUID = uuid.UUID("3bd2a492-96c0-4079-b420-fcf98ef103ed")
    # EFI_CERT_X509_SHA384_GUID = uuid.UUID("0x7076876e, 0x80c2, 0x4ee6, 0xaa, 0xd2, 0x28, 0xb3, 0x49, 0xa6, 0x86, 0x5b")     # noqa: E501
    # EFI_CERT_X509_SHA512_GUID = uuid.UUID("0x446dbf63, 0x2502, 0x4cda, 0xbc, 0xfa, 0x24, 0x65, 0xd2, 0xb0, 0xfe, 0x9d")     # noqa: E501
    # EFI_CERT_TYPE_PKCS7_GUID = uuid.UUID("0x4aafd29d, 0x68df, 0x49ee, 0x8a, 0xa9, 0x34, 0x7d, 0x37, 0x56, 0x65, 0xa7")

    #
    # This method is a factory for creating the correct Efi Signature Data object
    # from the filestream of an existing auth payload
    #
    @staticmethod
    def Factory(fs, type, size):
        if(fs is None):
            raise Exception("Invalid File stream")

        if(type == EfiSignatureDataFactory.EFI_CERT_SHA256_GUID):
            if(size != EfiSignatureDataEfiCertSha256.STATIC_STRUCT_SIZE):
                raise Exception("Invalid Size 0x%x" % size)
            return EfiSignatureDataEfiCertSha256(decodefs=fs)

        elif(type == EfiSignatureDataFactory.EFI_CERT_X509_GUID):
            return EfiSignatureDataEfiCertX509(decodefs=fs, decodesize=size)

        else:
            logging.error("GuidType Value: %s" % type)
            raise Exception("Not Supported")
            return None

    #
    # Create a new Efi Signature Data object.
    #   Type will be baed on GUID
    #   Value will be based on type and Content (content stream opened for reading)
    #   sigowner is the UUID object for the signature owner guid
    @staticmethod
    def Create(type, ContentFileStream, sigowner):
        if(ContentFileStream is None):
            raise Exception("Invalid Content File Stream")

        if(type == EfiSignatureDataFactory.EFI_CERT_SHA256_GUID):
            return EfiSignatureDataEfiCertSha256(createfs=ContentFileStream, sigowner=sigowner)
        elif(type == EfiSignatureDataFactory.EFI_CERT_X509_GUID):
            return EfiSignatureDataEfiCertX509(createfs=ContentFileStream, sigowner=sigowner)
        else:
            raise Exception("Not Supported")

##
# EFI_SIGNATURE_LIST structure
##


class EfiSignatureList(object):
    STATIC_STRUCT_SIZE = 16 + 4 + 4 + 4

    def __init__(self, filestream=None, typeguid=None):
        if(filestream is None):

            # Type of the signature. GUID signature types are defined in below.
            self.SignatureType = typeguid

            # Total size of the signature list, including this header.
            self.SignatureListSize = EfiSignatureList.STATIC_STRUCT_SIZE

            # Size of the signature header which precedes the array of signatures.
            self.SignatureHeaderSize = -1

            # Size of each signature.
            self.SignatureSize = 0

            # Header before the array of signatures. The format of this header is specified by the SignatureType.
            self.SignatureHeader = None

            # An array of signatures. Each signature is SignatureSize bytes in length.
            self.SignatureData_List = None

        else:
            self.PopulateFromFileStream(filestream)

    def PopulateFromFileStream(self, fs):
        if(fs is None):
            raise Exception("Invalid File Steam")

        # only populate from file stream those parts that are complete in the file stream
        start = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(start)

        if((end - start) < EfiSignatureList.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")

        self.SignatureType = uuid.UUID(bytes_le=fs.read(16))
        self.SignatureListSize = struct.unpack("<I", fs.read(4))[0]
        self.SignatureHeaderSize = struct.unpack("<I", fs.read(4))[0]
        self.SignatureSize = struct.unpack("<I", fs.read(4))[0]

        # check the total size of this is within the File
        if((end - start) < self.SignatureListSize):
            logging.debug("SignatureListSize 0x%x" % self.SignatureListSize)
            logging.debug("End - Start is 0x%x" % (end - start))
            raise Exception("Invalid File Stream.  Not enough file content to cover the Sig List Size")

        # check that structure is built correctly and there is room within the structure total size to read the header
        if((self.SignatureListSize - (fs.tell() - start)) < self.SignatureHeaderSize):
            raise Exception("Invalid Sig List.  Sizes not correct.  "
                            "SignatureHeaderSize extends beyond end of structure")

        # Signature Header is allowed to be nothing (size 0)
        self.SignatureHeader = None
        if(self.SignatureHeaderSize > 0):
            self.SignatureHeader = EfiSignatureHeader(fs, self.SignatureHeaderSize)

        if(((self.SignatureListSize - (fs.tell() - start)) % self.SignatureSize) != 0):
            raise Exception("Invalid Sig List.  Signature Data Array is not a valid size")

        self.SignatureData_List = []
        while((start + self.SignatureListSize) > fs.tell()):
            # double check that everything is adding up correctly.
            if((start + self.SignatureListSize - fs.tell() - self.SignatureSize) < 0):
                raise Exception("Invalid Signature List Processing.  Signature Data not correctly parsed!!")
            a = EfiSignatureDataFactory.Factory(fs, self.SignatureType, self.SignatureSize)
            self.SignatureData_List.append(a)

    def Print(self):
        print("EfiSignatureList")
        print("  Signature Type:        %s" % str(self.SignatureType))
        print("  Signature List Size:   0x%x" % self.SignatureListSize)
        print("  Signature Header Size: 0x%x" % self.SignatureHeaderSize)
        print("  Signature Size:        0x%x" % self.SignatureSize)
        if(self.SignatureHeader is not None):
            self.SignatureHeader.Print()
        else:
            print("  Signature Header:      NONE")

        for a in self.SignatureData_List:
            a.Print()

    def Write(self, fs):
        if(fs is None):
            raise Exception("Invalid File Output Stream")
        if((self.SignatureHeader is None) and (self.SignatureHeaderSize == -1)):
            raise Exception("Invalid object.  Uninitialized Sig Header")
        if(self.SignatureData_List is None):
            raise Exception("Invalid object.  No Sig Data")

        fs.write(self.SignatureType.bytes_le)
        fs.write(struct.pack("<I", self.SignatureListSize))
        fs.write(struct.pack("<I", self.SignatureHeaderSize))
        fs.write(struct.pack("<I", self.SignatureSize))
        if(self.SignatureHeader is not None):
            self.SignatureHeader.Write(fs)

        for a in self.SignatureData_List:
            a.Write(fs)

    def AddSignatureHeader(self, SigHeader, SigSize=0):
        if(self.SignatureHeader is not None):
            raise Exception("Signature Header already set")

        if(self.SignatureHeaderSize != -1):
            raise Exception("Signature Header already set (size)")

        if(self.SignatureSize != 0):
            raise Exception("Signature Size already set")

        if(self.SignatureData_List is not None):
            raise Exception("Signature Data List is already initialized")

        if(SigHeader is None) and (SigSize == 0):
            raise Exception("Invalid parameters.  Can't have no header and 0 Signature Size")

        self.SignatureHeader = SigHeader
        if(SigHeader is None):
            self.SignatureHeaderSize = 0
            self.SignatureSize = SigSize
        else:
            self.SignatureHeaderSize = SigHeader.GetTotalSize()
            self.SignatureSize = SigHeader.GetSizeOfSignatureDataEntry()
            self.SignatureListSize += self.SignatureHeaderSize

    def AddSignatureData(self, SigDataObject):
        if(self.SignatureSize == 0):
            raise Exception("Before adding Signature Data you must have set the Signature Size")

        if(self.SignatureSize != SigDataObject.GetTotalSize()):
            raise Exception("Can't add Signature Data of different size")

        if(self.SignatureData_List is None):
            self.SignatureData_List = []

        self.SignatureData_List.append(SigDataObject)
        self.SignatureListSize += self.SignatureSize


class EfiTime(object):
    STATIC_STRUCT_SIZE = 16

    def __init__(self, Time=datetime.datetime.now(), decodefs=None):

        if(decodefs is None):
            self.Time = Time
        else:
            self.PopulateFromFileStream(decodefs)

    def PopulateFromFileStream(self, fs):
        if(fs is None):
            raise Exception("Invalid File Steam")

        # only populate from file stream those parts that are complete in the file stream
        start = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(start)

        if((end - start) < EfiTime.STATIC_STRUCT_SIZE):  # size of the static structure data
            raise Exception("Invalid file stream size")

        Year = struct.unpack("<H", fs.read(2))[0]
        Month = struct.unpack("<B", fs.read(1))[0]
        Day = struct.unpack("<B", fs.read(1))[0]
        Hour = struct.unpack("<B", fs.read(1))[0]
        Minute = struct.unpack("<B", fs.read(1))[0]
        Second = struct.unpack("<B", fs.read(1))[0]
        fs.seek(1, 1)  # seek past pad1
        NanoSecond = struct.unpack("<I", fs.read(4))[0]
        TimeZone = struct.unpack("<h", fs.read(2))[0]
        Daylight = struct.unpack("<B", fs.read(1))[0]
        fs.seek(1, 1)  # seek past pad2

        self.Time = datetime.datetime(Year, Month, Day, Hour, Minute, Second, NanoSecond / 1000)
        logging.debug("I don't know how to deal with TimeZone or Daylight and I don't care at the moment")
        logging.debug("Timezone value is: 0x%x" % TimeZone)
        logging.debug("Daylight value is: 0x%X" % Daylight)

    def Print(self):
        print("EfiTime: %s" % datetime.datetime.strftime(self.Time, "%A, %B %d, %Y %I:%M%p"))

    def Write(self, fs):
        if(fs is None):
            raise Exception("Invalid File Output Stream")

        fs.write(struct.pack("<H", self.Time.year))
        fs.write(struct.pack("<B", self.Time.month))
        fs.write(struct.pack("<B", self.Time.day))
        fs.write(struct.pack("<B", self.Time.hour))
        fs.write(struct.pack("<B", self.Time.minute))
        fs.write(struct.pack("<B", self.Time.second))
        fs.write(struct.pack("<B", 0))  # Pad1
        fs.write(struct.pack("<I", 0))  # Nano Seconds
        fs.write(struct.pack("<h", 0))  # TimeZone
        fs.write(struct.pack("<B", 0))  # Daylight
        fs.write(struct.pack("<B", 0))  # Pad2


class EFiVariableAuthentication2(object):

    def __init__(self, Time=datetime.datetime.now(), decodefs=None):
        if(decodefs is None):
            self.EfiTime = EfiTime(Time=Time)
            self.AuthInfo = WinCertUefiGuid()
            self.Payload = None
            self.PayloadSize = 0
            self.SigListPayload = None
        else:
            self.PopulateFromFileStream(decodefs)

    def PopulateFromFileStream(self, fs):
        if(fs is None):
            raise Exception("Invalid File Steam")
        self.EfiTime = EfiTime(decodefs=fs)
        self.AuthInfo = WinCert.Factory(fs)
        self.Payload = None
        self.SigListPayload = None

        self.SetPayload(fs)

    def Print(self):
        print("EFiVariableAuthentication2")
        self.EfiTime.Print()
        self.AuthInfo.Print()
        print("-------------------- VARIABLE PAYLOAD --------------------")
        if(self.SigListPayload is not None):
            self.SigListPayload.Print()

        elif(self.Payload is not None):
            print("Raw Data: ")
            sdl = self.Payload.tolist()
            if(self.PayloadSize != len(sdl)):
                raise Exception("Invalid Payload Data Size vs Length of data")
            PrintByteList(sdl)

    def Write(self, fs):
        if(fs is None):
            raise Exception("Invalid File Output Stream")

        self.EfiTime.Write(fs)
        self.AuthInfo.Write(fs)
        if(self.Payload is not None):
            fs.write(self.Payload)

    def SetPayload(self, fs):
        if(fs is None):
            raise Exception("Invalid File Input Stream")

        # Find the payload size
        start = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(start)
        self.PayloadSize = end - start
        if(self.PayloadSize == 0):
            logging.debug("No Payload for this EfiVariableAuthenticated2 Object")
            return

        # read as siglist
        try:
            self.SigListPayload = EfiSignatureList(fs)
        except Exception as e:
            logging.debug("Exception Trying to parse SigList Payload.  \n%s" % str(e))

        # reset the file pointer
        fs.seek(start)
        self.Payload = memoryview(fs.read(self.PayloadSize))


'''
THESE ARE NOT SUPPORTED IN THE TOOL

typedef struct {
  ///
  /// The SHA256 hash of an X.509 certificate's To-Be-Signed contents.
  ///
  EFI_SHA256_HASH     ToBeSignedHash;
  ///
  /// The time that the certificate shall be considered to be revoked.
  ///
  EFI_TIME            TimeOfRevocation;
} EFI_CERT_X509_SHA256;

typedef struct {
  ///
  /// The SHA384 hash of an X.509 certificate's To-Be-Signed contents.
  ///
  EFI_SHA384_HASH     ToBeSignedHash;
  ///
  /// The time that the certificate shall be considered to be revoked.
  ///
  EFI_TIME            TimeOfRevocation;
} EFI_CERT_X509_SHA384;

typedef struct {
  ///
  /// The SHA512 hash of an X.509 certificate's To-Be-Signed contents.
  ///
  EFI_SHA512_HASH     ToBeSignedHash;
  ///
  /// The time that the certificate shall be considered to be revoked.
  ///
  EFI_TIME            TimeOfRevocation;
} EFI_CERT_X509_SHA512;
'''



# @file
# Module contains defintions and structures from the UefiMultiPhase header file.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

EFI_VARIABLE_NON_VOLATILE = 0x00000001
EFI_VARIABLE_BOOTSERVICE_ACCESS = 0x00000002
EFI_VARIABLE_RUNTIME_ACCESS = 0x00000004
EFI_VARIABLE_HARDWARE_ERROR_RECORD = 0x00000008
EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS = 0x00000010
EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS = 0x00000020
EFI_VARIABLE_APPEND_WRITE = 0x00000040



# @file locate_tools_test.py
# unit test for locate_tools module.  Only runs on Windows machines.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import sys
import os
import edk2toollib.windows.locate_tools as locate_tools


class LocateToolsTest(unittest.TestCase):

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_GetVsWherePath(self):
        # Gets VSWhere
        old_vs_path = locate_tools.GetVsWherePath()
        os.remove(old_vs_path)
        self.assertFalse(os.path.isfile(old_vs_path), "This should be deleted")
        vs_path = locate_tools.GetVsWherePath()
        self.assertTrue(os.path.isfile(vs_path), "This should be back")

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindWithVsWhere(self):
        # Finds something with VSWhere
        ret, star_prod = locate_tools.FindWithVsWhere()
        self.assertEqual(ret, 0, "Return code should be zero")
        self.assertNotEqual(star_prod, None, "We should have found this product")
        ret, bad_prod = locate_tools.FindWithVsWhere("bad_prod")
        self.assertEqual(ret, 0, "Return code should be zero")
        self.assertEqual(bad_prod, None, "We should not have found this product")

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindwithVsWhereVs2017(self):
        ret, star_prod = locate_tools.FindWithVsWhere(vs_version="vs2017")
        self.assertEqual(ret, 0, "Return code should be zero")
        # not checking the result as no need to depend on machine state

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindwithVsWhereVs2019(self):
        ret, star_prod = locate_tools.FindWithVsWhere(vs_version="vs2019")
        self.assertEqual(ret, 0, "Return code should be zero")
        # not checking the result as no need to depend on machine state

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindWithVsWhereVs2015(self):
        ret, star_prod = locate_tools.FindWithVsWhere(vs_version="vs2015")
        self.assertEqual(ret, 2, "Return code should be 2 as Vs2015 isn't supported")

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindWithVsWhereUnsupported(self):
        ret, star_prod = locate_tools.FindWithVsWhere(vs_version="vs4096")
        self.assertEqual(ret, 2, "Return code should be two as unsupported version")

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_QueryVcVariables(self):
        keys = ["VCINSTALLDIR", "WindowsSDKVersion"]
        try:
            results = locate_tools.QueryVcVariables(keys)
        except ValueError:
            self.fail("We shouldn't assert in the QueryVcVariables")

        self.assertIsNotNone(results["VCINSTALLDIR"])
        self.assertIsNotNone(results["WindowsSDKVersion"])

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindInf2CatToolInWinSdk(self):
        results = locate_tools.FindToolInWinSdk("inf2cat.exe")
        self.assertIsNotNone(results)
        self.assertTrue(os.path.isfile(results))

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindToolInWinSdk(self):
        results = locate_tools.FindToolInWinSdk("signtool.exe")
        self.assertIsNotNone(results)
        self.assertTrue(os.path.isfile(results))

        results = locate_tools.FindToolInWinSdk("this_tool_should_never_exist.exe")
        self.assertIsNone(results)

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_QueryVcVariablesWithNoValidProduct(self):
        keys = ["VCINSTALLDIR", "WindowsSDKVersion"]
        with self.assertRaises(ValueError):
            locate_tools.QueryVcVariables(keys, product="YouWontFindThis")

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_QueryVcVariablesWithVariableNotFound(self):
        keys = ["YouWontFindMe", "WindowsSDKVersion"]
        with self.assertRaises(ValueError):
            locate_tools.QueryVcVariables(keys)

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_FindToolInWinSdkWithNoValidProduct(self):
        results = locate_tools.FindToolInWinSdk("WontFind.exe", product="YouWontFindThis")
        self.assertIsNone(results)



# @file
# Module contains helper classes and functions to work with UEFI FVs.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import uuid
import struct
import sys

#
# UEFI GUIDs
#
EfiSystemNvDataFvGuid = uuid.UUID(fields=(0xFFF12B8D, 0x7696, 0x4C8B, 0xA9, 0x85, 0x2747075B4F50))


#
# UEFI #Defines
#
EFI_FVH_SIGNATURE = b"_FVH"


#
# EFI_FIRMWARE_VOLUME_HEADER
# Can parse or produce an EFI_FIRMWARE_VOLUME_HEADER structure/byte buffer.
#
# typedef struct {
#   UINT8                     ZeroVector[16];
#   EFI_GUID                  FileSystemGuid;
#   UINT64                    FvLength;
#   UINT32                    Signature;
#   EFI_FVB_ATTRIBUTES_2      Attributes;
#   UINT16                    HeaderLength;
#   UINT16                    Checksum;
#   UINT16                    ExtHeaderOffset;
#   UINT8                     Reserved[1];
#   UINT8                     Revision;
#   EFI_FV_BLOCK_MAP_ENTRY    BlockMap[1];
# } EFI_FIRMWARE_VOLUME_HEADER;
class EfiFirmwareVolumeHeader(object):
    def __init__(self):
        self.StructString = "=16s16sQ4sLHHHBBQQ"
        self.ZeroVector = None
        self.FileSystemGuid = None
        self.FvLength = None
        self.Attributes = None
        self.HeaderLength = None
        self.Checksum = None
        self.ExtHeaderOffset = None
        self.Reserved = None
        self.Revision = None
        self.Blockmap0 = None
        self.Blockmap1 = None

    def load_from_file(self, file):
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (self.ZeroVector, file_system_guid_bin, self.FvLength, self.Signature, self.Attributes,
            self.HeaderLength, self.Checksum, self.ExtHeaderOffset, self.Reserved, self.Revision,
            self.Blockmap0, self.Blockmap1) = struct.unpack(self.StructString, struct_bytes)

        # Make sure that this structure is what we think it is.
        if self.Signature != EFI_FVH_SIGNATURE:
            raise Exception("File does not appear to point to a valid EfiFirmwareVolumeHeader!")

        # Update the GUID to be a UUID object.
        if sys.byteorder == 'big':
            self.FileSystemGuid = uuid.UUID(bytes=file_system_guid_bin)
        else:
            self.FileSystemGuid = uuid.UUID(bytes_le=file_system_guid_bin)

        return self

    def serialize(self):
        file_system_guid_bin = self.FileSystemGuid.bytes if sys.byteorder == 'big' else self.FileSystemGuid.bytes_le
        return struct.pack(self.StructString, self.ZeroVector, file_system_guid_bin, self.FvLength, self.Signature,
                           self.Attributes, self.HeaderLength, self.Checksum, self.ExtHeaderOffset, self.Reserved,
                           self.Revision, self.Blockmap0, self.Blockmap1)
#
# EFI_FIRMWARE_VOLUME_EXT_HEADER
# Can parse or produce an EFI_FIRMWARE_VOLUME_EXT_HEADER structure/byte buffer.
#
# typedef struct {
#   EFI_GUID                  FileSystemGuid;
#   UINT32                    ExtHeaderSize;
# } EFI_FIRMWARE_VOLUME_EXT_HEADER;


class EfiFirmwareVolumeExtHeader(object):
    def __init__(self):
        self.StructString = "=16sL"
        self.FileSystemGuid = None
        self.ExtHeaderSize = None

    def load_from_file(self, file):
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (self.FileSystemGuid, self.ExtHeaderSize) = struct.unpack(self.StructString, struct_bytes)

        return self



## @file override_parser_test.py
# Contains unit test routines for the OverrideParser class.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import os

from edk2toollib.uefi.edk2.parsers.override_parser import OverrideParser

SAMPLE_DATA_SINGLE_OVERRIDE = """#Override : 00000001 | My/Path/1 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27T22-36-30"""  # noqa: E501

SAMPLE_DATA_BAD_VERSION = """#Override : 0000000X | My/Path/1 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27T22-36-30"""      # noqa: E501
SAMPLE_DATA_BAD_DATE = """#Override : 00000001 | My/Path/1 | 4e367990b327501d1ea6fbee4002f9c8 | NOTADATE"""

SAMPLE_DATA_TRIPLE_OVERRIDE = """#Override : 00000001 | My/Path/1 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27T22-36-30
#Override : 00000001 | My/Path/2 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27T22-36-30
#Override : 00000001 | My/Path/3 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27T22-36-30
"""

SAMPLE_DATA_MIXED_CASE_OVERRIDE = """#Override : 00000001 | My/Path/1 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27T22-36-30
#OVERRIDE : 00000001 | MY/PATH/2 | 4E367990B327501D1EA6FBEE4002F9C8 | 2018-11-27T22-36-30
#override : 00000001 | my/path/3 | 4e367990b327501d1ea6fbee4002f9c8 | 2018-11-27t22-36-30
"""

SAMPLE_DATA_NO_OVERRIDE = """This are just some lines of text.

Nothing to see here.

#OVER... not really, bet you thought it was an override.
"""

SAMPLE_DATA_REAL_WORLD = """## @file
# The DXE driver produces FORM DISPLAY ENGINE protocol.
#
# Copyright (c) 2007 - 2018, Intel Corporation. All rights reserved.<BR>
# Copyright (c) 2015 - 2018, Microsoft Corporation.
#
#  This program and the accompanying materials
#  are licensed and made available under the terms and conditions of the BSD License
#  which accompanies this distribution. The full text of the license may be found at
#  http://opensource.org/licenses/bsd-license.php
#
#  THE PROGRAM IS DISTRIBUTED UNDER THE BSD LICENSE ON AN "AS IS" BASIS,
#  WITHOUT WARRANTIES OR REPRESENTATIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED.
#
#
##

#Override : 00000001 | MdeModulePkg/Universal/DisplayEngineDxe/DisplayEngineDxe.inf | 2eba0bc48b8ab3b1399c26800d057102 | 2018-10-05T21-43-51

[Defines]
  INF_VERSION                    = 0x00010005
  BASE_NAME                      = DisplayEngine
  FILE_GUID                      = E660EA85-058E-4b55-A54B-F02F83A24707
  MODULE_TYPE                    = DXE_DRIVER
  VERSION_STRING                 = 1.0
  ENTRY_POINT                    = InitializeDisplayEngine
  UNLOAD_IMAGE                   = UnloadDisplayEngine

# MORE TEXT BELOW HERE...
"""  # noqa: E501


class TestOverrideParser(unittest.TestCase):

    def test_no_inputs_raises_error(self):
        with self.assertRaises(ValueError):
            OverrideParser()

    def test_no_overrides_raises_error(self):
        with self.assertRaises(ValueError):
            OverrideParser(inf_contents=SAMPLE_DATA_NO_OVERRIDE)

    def test_bad_version_or_bad_date_raises_error(self):
        with self.assertRaises(ValueError):
            OverrideParser(inf_contents=SAMPLE_DATA_BAD_VERSION)
        with self.assertRaises(ValueError):
            OverrideParser(inf_contents=SAMPLE_DATA_BAD_DATE)

    def test_single_override_is_parsed(self):
        parser = OverrideParser(inf_contents=SAMPLE_DATA_SINGLE_OVERRIDE)

        self.assertEqual(len(parser.override_lines), 1)
        self.assertEqual(len(parser.overrides), 1)

        self.assertEqual(parser.overrides[0]['version'], 1)
        self.assertEqual(parser.overrides[0]['original_path'].upper(), os.path.normpath('MY/PATH/1'))
        self.assertEqual(parser.overrides[0]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')
        self.assertEqual(parser.overrides[0]['datetime'].year, 2018)
        self.assertEqual(parser.overrides[0]['datetime'].month, 11)
        self.assertEqual(parser.overrides[0]['datetime'].day, 27)

    def test_real_world_override_is_parsed(self):
        parser = OverrideParser(inf_contents=SAMPLE_DATA_REAL_WORLD)

        self.assertEqual(len(parser.override_lines), 1)
        self.assertEqual(len(parser.overrides), 1)

        self.assertEqual(parser.overrides[0]['version'], 1)
        self.assertEqual(parser.overrides[0]['original_path'],
                         os.path.normpath('MdeModulePkg/Universal/DisplayEngineDxe/DisplayEngineDxe.inf'))
        self.assertEqual(parser.overrides[0]['current_hash'].upper(), '2EBA0BC48B8AB3B1399C26800D057102')
        self.assertEqual(parser.overrides[0]['datetime'].year, 2018)
        self.assertEqual(parser.overrides[0]['datetime'].month, 10)
        self.assertEqual(parser.overrides[0]['datetime'].day, 5)

    def test_triple_override_is_parsed(self):
        parser = OverrideParser(inf_contents=SAMPLE_DATA_TRIPLE_OVERRIDE)

        self.assertEqual(len(parser.override_lines), 3)
        self.assertEqual(len(parser.overrides), 3)

        self.assertEqual(parser.overrides[0]['version'], 1)
        self.assertEqual(parser.overrides[0]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')
        self.assertEqual(parser.overrides[1]['version'], 1)
        self.assertEqual(parser.overrides[1]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')
        self.assertEqual(parser.overrides[2]['version'], 1)
        self.assertEqual(parser.overrides[2]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')

    def test_mixed_case_override_is_parsed(self):
        parser = OverrideParser(inf_contents=SAMPLE_DATA_MIXED_CASE_OVERRIDE)

        self.assertEqual(len(parser.override_lines), 3)
        self.assertEqual(len(parser.overrides), 3)

        self.assertEqual(parser.overrides[0]['version'], 1)
        self.assertEqual(parser.overrides[0]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')
        self.assertEqual(parser.overrides[1]['version'], 1)
        self.assertEqual(parser.overrides[1]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')
        self.assertEqual(parser.overrides[2]['version'], 1)
        self.assertEqual(parser.overrides[2]['current_hash'].upper(), '4E367990B327501D1EA6FBEE4002F9C8')

    def test_parses_all_versions(self):
        # TODO: Fill out this test if it's ever needed.
        pass



# @file tpm2_stream_test.py
# This file contains utility classes to help marshal and unmarshal data to/from the TPM.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import unittest
import struct
from edk2toollib.tpm import tpm2_defs as Tpm2Defs
from edk2toollib.tpm import tpm2_stream as Tpm2Stream


class Tpm2StreamElement(unittest.TestCase):

    def test_object_has_zero_size_by_default(self):
        so = Tpm2Stream.Tpm2StreamElement()
        self.assertEqual(so.get_size(), 0)


class Tpm2CommandHeader(unittest.TestCase):

    def test_ch_marshals_correctly(self):
        ch1 = Tpm2Stream.TPM2_COMMAND_HEADER(0x4321, 0x00000000, 0xDEADBEEF)
        ch2 = Tpm2Stream.TPM2_COMMAND_HEADER(0x8001, 0x0000000A, Tpm2Defs.TPM_CC_Clear)

        self.assertEqual(ch1.marshal(), bytearray.fromhex('432100000000DEADBEEF'))
        self.assertEqual(ch2.marshal(), bytearray.fromhex('80010000000A') + struct.pack(">L", Tpm2Defs.TPM_CC_Clear))

    def test_ch_has_correct_size(self):
        ch1 = Tpm2Stream.TPM2_COMMAND_HEADER(0x4321, 0x00000000, 0xDEADBEEF)
        self.assertEqual(ch1.get_size(), 0x0A)

    def test_ch_size_can_be_updated(self):
        ch1 = Tpm2Stream.TPM2_COMMAND_HEADER(0x4321, 0x00000000, 0xDEADBEEF)
        self.assertEqual(ch1.marshal(), bytearray.fromhex('432100000000DEADBEEF'))
        ch1.update_size(0x1234)
        self.assertEqual(ch1.marshal(), bytearray.fromhex('432100001234DEADBEEF'))



## @file setup.py
# This contains setup info for edk2-pytool-library pip module
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import setuptools
from setuptools.command.sdist import sdist
from setuptools.command.install import install
from setuptools.command.develop import develop
from edk2toollib.windows.locate_tools import _DownloadVsWhere

with open("readme.md", "r") as fh:
    long_description = fh.read()


class PostSdistCommand(sdist):
    """Post-sdist."""
    def run(self):
        # we need to download vswhere so throw the exception if we don't get it
        _DownloadVsWhere()
        sdist.run(self)


class PostInstallCommand(install):
    """Post-install."""
    def run(self):
        install.run(self)
        _DownloadVsWhere()


class PostDevCommand(develop):
    """Post-develop."""
    def run(self):
        develop.run(self)
        try:
            _DownloadVsWhere()
        except:
            pass


setuptools.setup(
    name="edk2-pytool-library",
    author="Tianocore Edk2-PyTool-Library team",
    author_email="sean.brogan@microsoft.com",
    description="Python library supporting UEFI EDK2 firmware development",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tianocore/edk2-pytool-library",
    license='BSD-2-Clause-Patent',
    packages=setuptools.find_packages(),
    cmdclass={
        'sdist': PostSdistCommand,
        'install': PostInstallCommand,
        'develop': PostDevCommand,
    },
    include_package_data=True,
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers"
    ]
)



## @file
# Quick script to check that the wheel/package created is aligned on a git tag.
# Official releases should not be made from non-tagged code.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import glob
import os
import sys

p = os.path.join(os.getcwd(), "dist")
whlfile = glob.glob(os.path.join(p, "*.whl"))
if(len(whlfile) != 1):
    for filename in whlfile:
        print(filename)
    raise Exception("Too many wheel files")
rfn = os.path.relpath(whlfile[0], os.getcwd())
v = rfn.split("-")[1]
if v.count(".") != 2:
    raise Exception("Version %s not in format major.minor.patch" % v)
if "dev" in v:
    raise Exception("No Dev versions allowed to be published.")
print("version: " + str(v))
sys.exit(0)



# @file BaseParser.py
# Code to support parsing EDK2 files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import logging


class BaseParser(object):

    def __init__(self, log):
        self.Logger = logging.getLogger(log)
        self.Lines = []
        self.LocalVars = {}
        self.InputVars = {}
        self.CurrentSection = ""
        self.CurrentFullSection = ""
        self.Parsed = False
        self.ConditionalStack = []
        self.RootPath = ""
        self.PPs = []
        self.TargetFile = None
        self.TargetFilePath = None

    #
    # For include files set the base root path
    #

    def SetBaseAbsPath(self, path):
        self.RootPath = path
        return self

    def SetPackagePaths(self, pps=[]):
        self.PPs = pps
        return self

    def SetInputVars(self, inputdict):
        self.InputVars = inputdict
        return self

    def FindPath(self, *p):
        # NOTE: Some of this logic should be replaced
        #       with the path resolution from Edk2Module code.

        # If the absolute path exists, return it.
        Path = os.path.join(self.RootPath, *p)
        if os.path.exists(Path):
            return Path

        # If that fails, check a path relative to the target file.
        if self.TargetFilePath is not None:
            Path = os.path.join(self.TargetFilePath, *p)
            if os.path.exists(Path):
                return Path

        # If that fails, check in every possible Pkg path.
        for Pkg in self.PPs:
            Path = os.path.join(self.RootPath, Pkg, *p)
            if os.path.exists(Path):
                return Path

        # log invalid file path
        Path = os.path.join(self.RootPath, *p)
        self.Logger.error("Invalid file path %s" % Path)
        return Path

    def WriteLinesToFile(self, filepath):
        self.Logger.debug("Writing all lines to file: %s" % filepath)
        f = open(filepath, "w")
        for l in self.Lines:
            f.write(l + "\n")
        f.close()
    #
    # do logical comparisons
    #

    def ComputeResult(self, value, cond, value2):
        if(cond == "=="):
            # equal
            return (value.upper() == value2.upper())

        elif (cond == "!="):
            # not equal
            return (value.upper() != value2.upper())

        elif (cond == "<"):
            return (self.ConvertToInt(value) < (self.ConvertToInt(value2)))

        elif (cond == "<="):
            return (self.ConvertToInt(value) <= (self.ConvertToInt(value2)))

        elif (cond == ">"):
            return (self.ConvertToInt(value) > (self.ConvertToInt(value2)))

        elif (cond == ">="):
            return (self.ConvertToInt(value) >= (self.ConvertToInt(value2)))

    #
    # convert to int based on prefix
    #

    def ConvertToInt(self, value):
        if(value.upper().startswith("0X")):
            return int(value, 16)
        else:
            return int(value, 10)

    #
    # Push new value on stack
    #
    def PushConditional(self, v):
        self.ConditionalStack.append(v)

    #
    # Pop conditional and return the value
    #

    def PopConditional(self):
        if(len(self.ConditionalStack) > 0):
            return self.ConditionalStack.pop()
        else:
            self.Logger.critical("Tried to pop an empty conditional stack.  Line Number %d" % self.CurrentLine)
            return self.ConditionalStack.pop()  # this should cause a crash but will give trace.

    #
    # Method to replace variables
    # in a line with their value from input dict or local dict
    #
    def ReplaceVariables(self, line):
        rep = line.count("$")
        result = line
        index = 0
        while(rep > 0):
            start = line.find("$(", index)
            end = line.find(")", start)

            token = line[start + 2:end]
            retoken = line[start:end + 1]
            self.Logger.debug("Token is %s" % token)
            v = self.LocalVars.get(token)
            self.Logger.debug("Trying to replace %s" % retoken)
            if(v is not None):
                #
                # fixme: This should just be a workaround!!!!!
                #
                if (v.upper() == "TRUE" or v.upper() == "FALSE"):
                    v = v.upper()
                self.Logger.debug("with %s  [From Local Vars]" % v)
                result = result.replace(retoken, v, 1)
            else:
                # use the passed in Env
                v = self.InputVars.get(token)

                if(v is None):
                    self.Logger.error("Unknown variable %s in  %s" % (token, line))
                    # raise Exception("Invalid Variable Replacement", token)
                    # just skip it because we need to support ifdef
                else:
                    # found in the Env
                    #
                    # fixme: This should just be a workaround!!!!!
                    #
                    if (v.upper() == "TRUE" or v.upper() == "FALSE"):
                        v = v.upper()
                    self.Logger.debug("with %s [From Input Vars]" % v)
                    result = result.replace(retoken, v, 1)

            index = end + 1
            rep = rep - 1

        return result

    #
    # Process Conditional
    # return true if line is a conditional otherwise false
    #

    def ProcessConditional(self, text):
        tokens = text.split()
        if(tokens[0].lower() == "!if"):
            # need to add support for OR/AND
            if(len(tokens) < 4):
                self.Logger.error("!if conditionals need to be formatted correctly (spaces between each token)")
                raise Exception("Invalid conditional", text)
            con = self.ComputeResult(tokens[1].strip(), tokens[2].strip(), tokens[3].strip())
            self.PushConditional(con)
            return True

        elif(tokens[0].lower() == "!ifdef"):
            self.PushConditional((tokens[1].count("$") == 0))
            return True

        elif(tokens[0].lower() == "!ifndef"):
            self.PushConditional((tokens[1].count("$") > 0))
            return True

        elif(tokens[0].lower() == "!else"):
            v = self.PopConditional()
            self.PushConditional(not v)
            return True

        elif(tokens[0].lower() == "!endif"):
            self.PopConditional()
            return True

        return False

    #
    # returns true or false depending on what state of conditional you are currently in
    #
    def InActiveCode(self):
        ret = True
        for a in self.ConditionalStack:
            if not a:
                ret = False
                break

        return ret

    #
    # will return true if the the line has
    # { 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}
    #

    def IsGuidString(self, l):
        if(l.count("{") == 2 and l.count("}") == 2 and l.count(",") == 10 and l.count("=") == 1):
            return True
        return False

    def ParseGuid(self, l):
        # parse a guid in format
        # { 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}
        # into F7FDE4A6-294C-493c-B50F-9734553BB757  (NOTE these are not same guid this is just example of format)
        entries = l.lstrip(' {').rstrip(' }').split(',')
        gu = entries[0].lstrip(' 0').lstrip('x').strip()
        # pad front until 8 chars
        while(len(gu) < 8):
            gu = "0" + gu

        gut = entries[1].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 4):
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[2].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 4):
            gut = "0" + gut
        gu = gu + "-" + gut

        # strip off extra {
        gut = entries[3].lstrip(' { 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[4].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[5].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[6].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[7].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[8].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[9].lstrip(' 0').lstrip('x').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        gut = entries[10].split()[0].lstrip(' 0').lstrip('x').rstrip(' } ').strip()
        while(len(gut) < 2):
            gut = "0" + gut
        gu = gu + gut

        return gu.upper()

    def ResetParserState(self):
        self.ConditionalStack = []
        self.CurrentSection = ''
        self.CurrentFullSection = ''
        self.Parsed = False

#
# Base Class for Edk2 build files that use # for comments
#


class HashFileParser(BaseParser):

    def __init__(self, log):
        BaseParser.__init__(self, log)

    def StripComment(self, l):
        return l.split('#')[0].strip()

    def ParseNewSection(self, l):
        if(l.count("[") == 1 and l.count("]") == 1):  # new section
            section = l.strip().lstrip("[").split(".")[0].split(",")[0].rstrip("]").strip()
            self.CurrentFullSection = l.strip().lstrip("[").split(",")[0].rstrip("]").strip()
            return (True, section)
        return (False, "")



##
# Handle basic logging by streaming into stringIO
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import logging
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class StringStreamHandler(logging.StreamHandler):
    terminator = '\n'

    def __init__(self):
        logging.Handler.__init__(self)
        self.stream = StringIO()

    def handle(self, record):
        """
        Conditionally emit the specified logging record.
        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """

        rv = self.filter(record)
        if rv and record.levelno >= self.level:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv

    def readlines(self, hint=-1):
        return self.stream.readlines(hint)

    def seek_start(self):
        self.stream.seek(0, 0)

    def seek_end(self):
        self.stream.seek(2, 0)

    def seek(self, offset, whence):
        return self.stream.seek(offset, whence)
